"""Akrocansim is a CAN bus J1939 controller simulator."""

__version__ = '0.2.0'

from pathlib import Path
import tomllib
import pickle
import canbustransmitter
import J1939
import gui
import can


akrocansim_dir = 'akrocansim'

default_config_toml = '''[CAN_INTERFACE]
# Akrocansim uses the python-can library for utilising CAN interfaces.
# Any configuration parameters defined by python-can, can be listed here
# in the same format at they would appear in can.interface.Bus().
# See https://python-can.readthedocs.io/en/stable/configuration.html.
# See example key/value pairs below for PCAN-USB.
interface='pcan'
channel='PCAN_USBBUS1'
bitrate=250000

[J1939DA]
# Obtain a copy of the J1939 Digital Annex from SAE International at: https://www.sae.org/standards/?search=j1939DA
# The J1939 Digital Annex is a Microsoft Excel file.
# If your copy is in .xls format, you need to convert it to .xlsx using Microsoft Excel.
# Save the 'J1939DA_??????.xlsx' file in the 'J1939' folder located in the same folder with this configuration.
# Edit the values of all keys in the [J1939DA] table as required and run J1939DA_parser.py.
filename = 'J1939DA_??????.xlsx'
SPNs_and_PGNs_sheet = 'SPNs & PGNs'

[J1939DA.SPNs_and_PGNs_sheet_columns]
# Adjust as required
'PGN' = 'E'
'Parameter Group Label' = 'F'
'Acronym' = 'G'
'PGN Description' = 'H'
'PGN Data Length' = 'O'
'Default Priority' = 'P'
'Transmission Rate' = 'N'
'SPN' = 'S'
'SPN Name' = 'T'
'SPN Description' = 'U'
'SPN Position in PGN' = 'R'
'SPN Length' = 'V'
'Resolution' = 'W'
'Offset' = 'X'
'Data Range' = 'Y'
'Operational Range' = 'Z'
'Units' = 'AA'

[J1939DA.SPNs_to_parse]
# Specify the range of rows in the J1939 Digital Annex SPNs_and_PGNs_sheet that will be scanned for SPN definitions.
first_row = 5
last_row = 5000

[Tx_PGNs_SPNs]
# List the PGNs and SPNs to be loaded by Akrocansim using the following format:
# PGN = [SPN#1, SPN#2, ..., SPN#N], e.g. 61444 = [513, 190]
#61444 = [513, 190]
'''


config_file = Path.home().joinpath(akrocansim_dir, 'config.toml')
config_file.parent.mkdir(exist_ok=True)
Path.home().joinpath(akrocansim_dir, 'J1939').mkdir(exist_ok=True)


if not config_file.exists():
    with config_file.open('w', encoding='utf-8') as f:
        f.write(default_config_toml)
        print(f'Configuration file created at: {config_file}\n'
              f'Follow the instructions in the configuration file and run Akrocansim again.')
else:
    with config_file.open('rb') as f:
        config = tomllib.load(f)

    Tx_PGNs_SPNs = {}
    for pgn, spn_list in config['Tx_PGNs_SPNs'].items():
        Tx_PGNs_SPNs[int(pgn)] = spn_list
    gui.Tx_PGNs_SPNs = Tx_PGNs_SPNs

    if config['CAN_INTERFACE']:
        canbustransmitter.bus = can.thread_safe_bus.Bus(**config['CAN_INTERFACE'])


J1939_pickle = Path.home().joinpath(akrocansim_dir, 'J1939', 'J1939.pkl')
if J1939_pickle.exists():
    with J1939_pickle.open('rb') as f:
        J1939_dict = pickle.load(f)
    canbustransmitter.J1939 = J1939_dict
    J1939.J1939 = J1939_dict
    gui.J1939 = J1939_dict
    gui.start_gui(canbustransmitter.bus, str(config_file))

    print(f'Using {J1939_pickle}')



import threading
import time
import can


bus: can.BusABC = None
J1939 = {}
PGNs_pending_tx = {}


def add_pending_tx(sender, app_data, pgn):
    global PGNs_pending_tx
    if pgn == 'all PGNs':
        PGNs_pending_tx = dict.fromkeys(PGNs_pending_tx, True)
    else:
        PGNs_pending_tx[pgn] = True


def start_pgn_tx(*, pgn, priority, source_address, tx_rate_ms, spns: list[int], get_value_fn):
    can_id = priority << 26 | pgn << 8 | source_address
    tx_rate_s = tx_rate_ms / 1000

    PGNs_pending_tx[pgn] = False

    data = [0 for _ in range(J1939[pgn]['PGN Data Length'])]

    def tx_pgn():
        while True:
            time.sleep(tx_rate_s)

            match get_value_fn('mode'), get_value_fn(f'{pgn} continuous tx'), PGNs_pending_tx[pgn]:
                case ['Tx All', _, _] | ['Use PGN Settings', True, _]:
                    # Transmit PGN
                    pass
                case [_, _, True]:
                    # Transmit PGN
                    PGNs_pending_tx[pgn] = False
                case ['Use PGN Settings', False, False] | ['Stop All', _, False]:
                    # Do not transmit PGN
                    continue
                case _:
                    raise ValueError('Unknown tx arbitration error')

            for spn in spns:
                spn_raw_value_tag = f'{spn}_raw_value'
                spn_spec = J1939[pgn]['SPNs'][spn]
                start_byte = spn_spec['start_byte']
                match spn_spec['length_bits']:
                    case 1 | 2 | 3 | 4 | 5 | 6 | 7:
                        start_bit = spn_spec['start_bit']
                        length_bits = spn_spec['length_bits']
                        mask = 0x00
                        for i in range(length_bits):
                            mask = mask | 1 << start_bit + i
                        data[start_byte] = data[start_byte] & ~mask | get_value_fn(spn_raw_value_tag) << start_bit & mask
                    case 8:
                        data[start_byte] = get_value_fn(spn_raw_value_tag)
                    case 16:
                        data[start_byte] = get_value_fn(spn_raw_value_tag) & 0x00FF
                        data[start_byte + 1] = get_value_fn(spn_raw_value_tag) >> 8
                    case 32:
                        data[start_byte] = get_value_fn(spn_raw_value_tag) & 0x0000_00FF
                        data[start_byte + 1] = get_value_fn(spn_raw_value_tag) >> 8 & 0x00_00FF
                        data[start_byte + 2] = get_value_fn(spn_raw_value_tag) >> 16 & 0x00FF
                        data[start_byte + 3] = get_value_fn(spn_raw_value_tag) >> 24

            if bus is not None:
                bus.send(can.Message(arbitration_id=can_id, is_extended_id=True, data=data))
            else:
                print('CAN ID:', f'{can_id:08X}', '  DATA:', data)

    threading.Thread(target=tx_pgn, daemon=True).start()



import os
import platform

import dearpygui.dearpygui as dpg
import canbustransmitter
import can
from J1939 import J1939, decode, encode, get_label, get_label_value


J1939 = {}
Tx_PGNs_SPNs = {}

def set_signal_value(sender, new_value, old_value):
    pgn, spn, sender_type = sender.split('_')
    pgn = int(pgn)
    spn = int(spn)
    if sender_type == 'input':
        dpg.set_value(f'{spn}_raw_value', encode(decoded_value=new_value,
                                                 scale=J1939[pgn]['SPNs'][spn]['scale'],
                                                 offset=J1939[pgn]['SPNs'][spn]['offset']))
    elif sender_type == 'combo':
        dpg.set_value(f'{spn}_raw_value', get_label_value(pgn=pgn, spn=spn, label=new_value))
    #if new_value > old_value:
    #    ...
    #else:
    #    ...
    #    dpg.set_value(spn, dpg.get_value(spn) - 1)
    #    dpg.set_item_user_data(sender, new_value)

def set_label(spn, new_value, pgn):
    dpg.set_value(f'{pgn}_{spn}_combo', get_label(pgn=pgn, spn=spn, raw_value=new_value))

def slider_min_value(pgn, spn):
    min_value = encode(decoded_value=J1939[pgn]['SPNs'][spn]['min_value'],
                       offset=J1939[pgn]['SPNs'][spn]['offset'],
                       scale=J1939[pgn]['SPNs'][spn]['scale'])
    return min_value

def slider_max_value(pgn, spn):
    max_value = encode(decoded_value=J1939[pgn]['SPNs'][spn]['max_value'],
                       offset=J1939[pgn]['SPNs'][spn]['offset'],
                       scale=J1939[pgn]['SPNs'][spn]['scale'])
    #(J1939[pgn]['SPNs'][spn]['max_value'] - J1939[pgn]['SPNs'][spn]['min_value']) / J1939[pgn]['SPNs'][spn]['scale'])
    if J1939[pgn]['SPNs'][spn]['length_bits'] < 32 or spn in [584, 585]: ###########################################################
        return max_value
    else:
        print('spn_max_value override [100_000]') ######################################################################################
        return 100_000 / J1939[pgn]['SPNs'][spn]['scale']


def add_pgn(pgn, spns: list):
    priority = J1939[pgn]['Default Priority']
    source_address = 0

    pgn_label = (f"PGN {pgn} - CAN ID: {priority << 26 | pgn << 8 | source_address:08X} "
                 f"- {J1939[pgn]['Acronym']} - {J1939[pgn]['Parameter Group Label']}")
    with (dpg.collapsing_header(label=pgn_label, default_open=True)):

        continuous_signals_present = False
        enum_signals_present = False

        for spn in spns:
            if J1939[pgn]['SPNs'][spn]['scale'] not in ['ENUM']:
                continuous_signals_present = True
            if J1939[pgn]['SPNs'][spn]['scale'] == 'ENUM':
                enum_signals_present = True

        if continuous_signals_present:
            with dpg.table(header_row=True, resizable=True,
                           borders_outerV=True, borders_outerH=True, borders_innerH=True, borders_innerV=True):
                dpg.add_table_column(label='SIGNAL', width_fixed=True)
                dpg.add_table_column(label='UNIT', width_fixed=True)
                dpg.add_table_column(label='RAW HEX', width_fixed=True)
                dpg.add_table_column(label='RAW DECIMAL')

                for spn in spns:
                    if J1939[pgn]['SPNs'][spn]['scale'] not in ['ENUM']:
                        with dpg.table_row():
                            # SIGNAL - decoded value, direct entry, decrement, decrement
                            max_value = (2 ** J1939[pgn]['SPNs'][spn]['length_bits'] - 1) \
                                        * J1939[pgn]['SPNs'][spn]['scale'] \
                                        + J1939[pgn]['SPNs'][spn]['offset']
                            if type(J1939[pgn]['SPNs'][spn]['scale']) is int:
                                dpg.add_input_int(tag=f'{pgn}_{spn}_input', width=140,
                                                  min_value=J1939[pgn]['SPNs'][spn]['offset'],
                                                  max_value=max_value,
                                                  min_clamped=True, max_clamped=True,
                                                  step=J1939[pgn]['SPNs'][spn]['scale'],
                                                  callback=set_signal_value, on_enter=True)
                            elif type(J1939[pgn]['SPNs'][spn]['scale']) is float:
                                dpg.add_input_float(tag=f'{pgn}_{spn}_input', width=140,
                                                    format=f"%.{J1939[pgn]['SPNs'][spn]['n_decimals']}f",
                                                    min_value=J1939[pgn]['SPNs'][spn]['offset'],
                                                    max_value=max_value,
                                                    min_clamped=True, max_clamped=True,
                                                    step=J1939[pgn]['SPNs'][spn]['scale'],
                                                    callback=set_signal_value, on_enter=True)
                            else:
                                raise TypeError

                            # UNIT
                            unit = J1939[pgn]['SPNs'][spn]['unit']
                            dpg.add_text(f"{unit if unit is not None else ''}".ljust(10))

                            # RAW HEX
                            dpg.add_text(tag=f'{spn}_hex')

                            # RAW DECIMAL
                            dpg.add_slider_int(tag=f'{spn}_raw_value', user_data=pgn,
                                               label=f"SPN {spn}: {J1939[pgn]['SPNs'][spn]['SPN Name']}",
                                               min_value=slider_min_value(pgn, spn),
                                               max_value=slider_max_value(pgn, spn),
                                               default_value=slider_min_value(pgn, spn))

        if enum_signals_present:
            with dpg.table(header_row=True, resizable=True,
                           borders_outerV=True, borders_outerH=True, borders_innerH=True, borders_innerV=True):
                dpg.add_table_column(label='DECIMAL', width_fixed=True)
                dpg.add_table_column(label='BINARY', width_fixed=True)
                dpg.add_table_column(label='HEX', width_fixed=True)
                dpg.add_table_column(label='OPTIONS')

                for spn in spns:
                    if J1939[pgn]['SPNs'][spn]['scale'] == 'ENUM':

                        with dpg.table_row():

                            # DECIMAL
                            dpg.add_input_int(tag=f'{spn}_raw_value', width=140, user_data=pgn, callback=set_label, on_enter=True,
                                              min_value=0, max_value=2 ** J1939[pgn]['SPNs'][spn]['length_bits'] - 1,
                                              min_clamped=True, max_clamped=True)

                            # BINARY
                            dpg.add_text(tag=f'{spn}_bin')

                            # HEX
                            dpg.add_text(tag=f'{spn}_hex')

                            # OPTIONS
                            labels = [label for label in J1939[pgn]['SPNs'][spn]['discrete_values'].values()]
                            dpg.add_combo(tag=f'{pgn}_{spn}_combo', items=labels,
                                          default_value=get_label(pgn=pgn, spn=spn, raw_value=0),
                                          label=f"SPN {spn}: {J1939[pgn]['SPNs'][spn]['SPN Name']}",
                                          callback=set_signal_value)

        with dpg.group(horizontal=True):
            dpg.add_text('Tx mode:')
            dpg.add_checkbox(tag=f'{pgn} continuous tx', label='Continuous')
            dpg.add_button(label='Tx Once', user_data=pgn, callback=canbustransmitter.add_pending_tx)

        dpg.add_spacer(height=10)

    canbustransmitter.start_pgn_tx(pgn=pgn, priority=priority, source_address=source_address,
                                   tx_rate_ms=J1939[pgn]['transmission_rate_ms'], spns=spns, get_value_fn=dpg.get_value)


def update_widgets():
    for pgn, spns in Tx_PGNs_SPNs.items():
        for spn in spns:
            raw_value = dpg.get_value(f'{spn}_raw_value')
            spn_spec = J1939[pgn]['SPNs'][spn]

            if spn_spec['scale'] == 'ENUM':
                pass
            else:
                decoded_value = decode(raw_value=raw_value,
                                       scale=spn_spec['scale'],
                                       offset=spn_spec['offset'])
                # :4.{len(str(J1939[pgn]['SPNs'][spn]['scale']).split('.')[1])}f
                dpg.set_value(f'{pgn}_{spn}_input', decoded_value)
                #dpg.set_item_user_data(f'{pgn}_{spn}_input', decoded_value)

            match spn_spec['length_bits'], spn_spec['scale']:
                case 1 | 2 | 3 | 4 as length_bits, 'ENUM':
                    nibble = f'{raw_value:X}'
                    dpg.set_value(f'{spn}_bin', f"{raw_value:0{length_bits}b}".ljust(10))
                    dpg.set_value(f'{spn}_hex', f'    [{nibble}]            ')
                case 5 | 6 | 7 | 8 as length_bits, 'ENUM':
                    byte = f'{raw_value:02X}'
                    dpg.set_value(f'{spn}_bin', f"{raw_value:0{length_bits}b}".ljust(10))
                    dpg.set_value(f'{spn}_hex', f'   [{byte}]            ')
                case 1 | 2 | 3 | 4, scale if scale not in ['ENUM']:
                    nimble = f'{raw_value:X}'
                    dpg.set_value(f'{spn}_hex', f'   [{nimble}]             ')
                case 5 | 6 | 7 | 8, scale if scale not in ['ENUM']:
                    byte = f'{raw_value:02X}'
                    dpg.set_value(f'{spn}_hex', f'   [{byte}]            ')
                case 16, _:
                    all_bytes = f'{raw_value:04X}'
                    byte1 = all_bytes[2:4]
                    byte2 = all_bytes[0:2]
                    dpg.set_value(f'{spn}_hex', f'LSB[{byte1} {byte2}]MSB      ')
                case 32, _:
                    all_bytes = f'{raw_value:08X}'
                    byte1 = all_bytes[6:8]
                    byte2 = all_bytes[4:6]
                    byte3 = all_bytes[2:4]
                    byte4 = all_bytes[0:2]
                    dpg.set_value(f'{spn}_hex', f'LSB[{byte1} {byte2} {byte3} {byte4}]MSB')


def start_gui(bus: can.BusABC, config_file: str):
    window_width = 1200
    window_height = 600
    dpg.create_context()
    dpg.create_viewport(title='Akro CAN Simulator', width=window_width, height=window_height, vsync=True)

    with dpg.window(label=bus.channel_info, width=window_width - 16, no_close=True):
        with dpg.menu_bar():
            with dpg.menu(label="Settings"):
                dpg.add_menu_item(label='Edit Configuration',
                                  callback=lambda s, a, u: os.system(config_file) if platform.system() == 'Windows'
                                  else os.system('%s %s' % (os.getenv('EDITOR'), config_file)))

        with dpg.group(horizontal=True):
            dpg.add_text('Continuous J1939 PGN transmission:')
            dpg.add_radio_button(tag='mode', items=('Stop All', 'Tx All', 'Use PGN Settings'),
                                 default_value='Stop All', horizontal=True)
        dpg.add_button(label='Tx All PGNs Once', user_data='all PGNs',
                       callback=canbustransmitter.add_pending_tx)

    with dpg.window(label='J1939', width=window_width - 16, pos=(0, 100), height=500, no_close=True):
        for pgn, spns in Tx_PGNs_SPNs.items():
            add_pgn(pgn, spns)

    dpg.setup_dearpygui()
    dpg.show_viewport()

    while dpg.is_dearpygui_running():
        update_widgets()
        dpg.render_dearpygui_frame()

    dpg.destroy_context()

    if bus is not None:
        bus.shutdown()



J1939 = {}

def decode(*, raw_value, scale, offset):
    return raw_value * scale + offset

def encode(*, decoded_value, scale, offset):
    return round((decoded_value - offset) / scale)

def get_label(*, pgn, spn, raw_value):
    try:
        return J1939[pgn]['SPNs'][spn]['discrete_values'][raw_value]
    except KeyError:
        return ''

def get_label_value(*, pgn, spn, label):
    for value, _label in J1939[pgn]['SPNs'][spn]['discrete_values'].items():
        if label == _label:
            return value

if __name__ == '__main__':
    assert -40 == decode(raw_value=0, scale=0.5, offset=-40)
    assert -40 == decode(raw_value=0, scale=1, offset=-40)
    assert -40 == decode(raw_value=0, scale=2, offset=-40)
    assert -20 == decode(raw_value=40, scale=0.5, offset=-40)
    assert 0 == decode(raw_value=40, scale=1, offset=-40)
    assert 40 == decode(raw_value=40, scale=2, offset=-40)
    assert -38 == decode(raw_value=1, scale=2, offset=-40)

    assert 0 == encode(decoded_value=-40, scale=0.5, offset=-40)
    assert 0 == encode(decoded_value=-40, scale=1, offset=-40)
    assert 0 == encode(decoded_value=-40, scale=2, offset=-40)
    assert 40 == encode(decoded_value=-20, scale=0.5, offset=-40)
    assert 40 == encode(decoded_value=0, scale=1, offset=-40)
    assert 40 == encode(decoded_value=40, scale=2, offset=-40)
    assert 1 == encode(decoded_value=-38, scale=2, offset=-40)
